// gsgc.c -- a generation scavenging garbage collector for C and C++ */

/* Copyright (c) 2011 Ian Piumarta
** All Rights Reserved
** 
** Permission is hereby granted, free of charge, to any person obtaining a
** copy of this software and associated documentation files (the 'Software'),
** to deal in the Software without restriction, including without limitation
** the rights to use, copy, modify, merge, publish, distribute, and/or sell
** copies of the Software, and to permit persons to whom the Software is
** furnished to do so, provided that the above copyright notice(s) and this
** permission notice appear in all copies of the Software.  Inclusion of the
** the above copyright notice(s) and this permission notice in supporting
** documentation would be appreciated but is not required.
** 
** THE SOFTWARE IS PROVIDED 'AS IS'.  USE ENTIRELY AT YOUR OWN RISK.
*/

// Last edited: 2011-09-16 07:08:34 by piumarta on vps2.piumarta.com

#if !defined(GC_MAX_AGE)
# define GC_MAX_AGE		4
#endif

#if !defined(GC_MAX_REMEMBERED)
# define GC_MAX_REMEMBERED	1024
#endif

#if !defined(GC_MIN_NEW_SPACE)
# define GC_MIN_NEW_SPACE	4*1024*1024
#endif

#if !defined(GC_ALIGN)
# define GC_ALIGN		sizeof(void *)
#endif

#if !defined(GC_IS_POINTER)
# define GC_IS_POINTER(PTR)	(PTR)
#endif

#include <stdio.h>

#include "gsgc.h"

#include <stdlib.h>
#include <string.h>
#include <assert.h>

#if (GC_STATS)
# include <sys/time.h>
#endif

struct Header {
  size_t	size;
  GC_PTR	forwardingPointer;
  long		map;
  char		age;
  char		remembered;
  char		old;
#if defined(GC_USER)
  GC_USER	user;
#endif
  GC_PTR	contents[0];
};

typedef struct Header Header;

static inline GC_PTR  hdr2ptr(Header *hdr)		{ return (GC_PTR)(hdr + 1); }
static inline Header *ptr2hdr(GC_PTR ptr)		{ return ((Header *)ptr) - 1; }
static inline Header *hdrNext(Header *hdr, int size)	{ return (Header *)((void *)hdr + size); }

GC_API int  GC_size(GC_PTR ptr)				{ return ptr2hdr(ptr)->size - sizeof(Header); }
GC_API long GC_is_pointers(GC_PTR ptr)			{ return ptr2hdr(ptr)->map; }

struct Space {
  const char	*name;	/* name of space */
  Header	*base;	/* first address in space */
  Header	*next;	/* next free address in space */
  Header	*limit;	/* first address beyond space */
};

typedef struct Space Space;

static void Space_init(struct Space *space, size_t size, const char *name)
{
  space->name=  name;
  space->base=  (Header *)malloc(size);
  space->next=  space->base;
  space->limit= hdrNext(space->base, size);
  bzero(space->base, size);
}

static void Space_grow(struct Space *space, size_t size)
{
  free(space->base);
  space->base=  (Header *)malloc(size);
  space->next=  space->base;
  space->limit= hdrNext(space->base, size);
}

static size_t Space_size(struct Space *space)			{ return (void *)space->limit - (void *)space->base; }
static size_t Space_used(struct Space *space)			{ return (void *)space->next  - (void *)space->base; }
static size_t Space_free(struct Space *space)			{ return (void *)space->limit - (void *)space->next; }

#if !defined(NDEBUG)
static int Space_includes(struct Space *space, void *ptr)	{ return (void *)space->base <= ptr && ptr < (void *)space->next; }
#endif

struct Table {
  void		**at;
  size_t	  count;
  size_t	  size;
};

typedef struct Table Table;

static Table *Table_init(Table *t, size_t size)
{
  t->at    = (void **)malloc(sizeof(void *) * size);
  t->count = 0;
  t->size  = size;
  return t;
}

static void Table_grow(Table *t)
{
  t->size *= 2;
  t->at= (void **)realloc(t->at, sizeof(void *) * t->size);
}

static inline void *Table_add(Table *t, void *ptr)
{
  assert(t->count <= t->size);
  if (t->count == t->size) Table_grow(t);
  assert(t->count < t->size);
  return t->at[t->count++]= ptr;
}

#if (GC_SPY)
# include "window.c"
static int spyCount= 3333;
GC_API void GC_spy_every(int n) { spyCount= n; }
#endif

struct GarbageCollector {
  Space	newSpace, survivorSpace, oldSpace, oldSurvivorSpace;
  Table	rememberedSet, roots, stack;
  int   seq, full;
//int   count;
#if 0
  GC_Callback callback;
#endif
#if (GC_STATS)
  int	totalAge, totalNew;
  struct timeval gcTime, startTime;
#endif
#if (GC_SPY)
  window *w;
#endif
};

#if (GC_SPY)

#define W1 300
#define W2 100

GC_API void GC_spy(GarbageCollector *gc, int flag)
{
  if (flag && !gc->w) gc->w= window_new(W1+W2+1, 40);
}

static int wd(Space *spc, int w)	{ return (Space_used(spc) >> 10) * w / (Space_size(spc) >> 10); }

static void GC_update_spy(GarbageCollector *gc)
{
  if (gc->w && List_isEmpty(&gc->w->requests.list)) {
    window_clear(gc->w);
    window_setColour(gc->w, 0.5, 0.5, 0.5);
    window_fillRect(gc->w,     1 , 31, wd(&gc->oldSurvivorSpace, W1-2), 8);
    window_fillRect(gc->w, (W1+1), 31, wd(&gc->survivorSpace   , W2-2), 8);
    window_setColour(gc->w, 0, 0, 0);
    window_fillRect(gc->w,     1 , 21, wd(&gc->oldSpace        , W1-2), 8);
    window_fillRect(gc->w, (W1+1), 21, wd(&gc->newSpace        , W2-2), 8);
    window_drawRect(gc->w,     1 , 31,                           W1-2,  8);
    window_drawRect(gc->w, (W1+1), 31,                           W2-2,  8);
    window_drawRect(gc->w,     1 , 21,                           W1-2,  8);
    window_drawRect(gc->w, (W1+1), 21,                           W2-2,  8);
    char buf[80];
    snprintf(buf, sizeof(buf), "roots: %2i+%2i/%5i stack: %4i MB: %ld = %7ld+%ld",
	     (int)gc->roots.count, (int)gc->rememberedSet.count, (int)gc->rememberedSet.size, (int)gc->stack.count,
	     (long)Space_size(&gc->oldSpace) / 1024 + (long)Space_size(&gc->newSpace) / 1024,
	     (long)Space_size(&gc->oldSpace) / 1024, (long)Space_size(&gc->newSpace) / 1024);
    window_text(gc->w, 2, 6, buf);
    window_flush(gc->w);
  }
}

#endif

#if (GC_DEBUG)
  static void sanityCheck(GarbageCollector *gc);
#else
# define sanityCheck(GC)
#endif

#if 0
static void nullCallback(GarbageCollector *gc) {}

GC_API void GC_set_callback(GarbageCollector *gc, GC_Callback function)
{
  gc->callback= function;
}
#endif

GarbageCollector *GC_init(GarbageCollector *gc, size_t size)
{
  size= ((size + (GC_ALIGN - 1)) & -GC_ALIGN);
  if (size < GC_MIN_NEW_SPACE) size= GC_MIN_NEW_SPACE;
  Space_init(&gc->oldSpace,		size * 2, "oldSpace");
  Space_init(&gc->oldSurvivorSpace,	size * 2, "oldSSpace");
  Space_init(&gc->newSpace,		size, "newSpace");
  Space_init(&gc->survivorSpace,	size, "newSSpace");
  Table_init(&gc->rememberedSet,	GC_MAX_REMEMBERED);
  Table_init(&gc->roots, 32);
  Table_init(&gc->stack, 32);
  gc->seq= 0;
  gc->full= 0;
//gc->count= 0;
#if 0
  gc->callback= nullCallback;
#endif
#if (GC_STATS)
  timerclear(&gc->startTime);
  timerclear(&gc->gcTime);
  gettimeofday(&gc->startTime, 0);
#endif
  return gc;
}

GC_API void GC_delete(GarbageCollector *gc)
{
#if (GC_STATS)
  struct timeval now;
  gettimeofday(&now, 0);
  timersub(&now, &gc->startTime, &now);
  long msGC= gc->gcTime.tv_sec * 1000 + gc->gcTime.tv_usec / 1000;
  long msTotal= now.tv_sec * 1000 + now.tv_usec / 1000;
  printf("\n[GC overhead ");
  printf("%ld/%ldms", msGC, msTotal);
  if (msTotal) printf(", %4.1f%% of total execution time", (float)msGC * 100.0 / (float)msTotal);
  printf("]\n");
#endif
}

GC_API GarbageCollector *GC_new(size_t size)
{
  return GC_init((GarbageCollector *)malloc(sizeof(GarbageCollector)), size);
}

GC_API inline void GC_add_root(GarbageCollector *gc, GC_PTR *root)
{
  Table_add(&gc->roots, root);
}

GC_API inline GC_StackMarker GC_begin(GarbageCollector *gc)
{
  return gc->stack.count;
}

GC_API inline void GC_push_root(GarbageCollector *gc, GC_PTR *root, char *memo)
{
  Table_add(&gc->stack, root);
}

GC_API inline void GC_end(GarbageCollector *gc, GC_StackMarker mark)
{
  assert(mark <= gc->stack.count);
  gc->stack.count= mark;
}

static void remember(GarbageCollector *gc, GC_PTR ptr)
{
  assert(ptr2hdr(ptr)->old);
  ptr2hdr(ptr)->remembered= 1;
  Table_add(&gc->rememberedSet, ptr);
}

static Header *copyToSpace(Header *old, struct Space *space)
{
  int     size= old->size;
  Header *new=  space->next;
  Header *next= hdrNext(new, size);
  if (next >= space->limit) {
    fprintf(stderr, "%s is full", space->name);
    abort();
  }
  memcpy(new, old, size);
  space->next= next;
  return new;
}

static void copyAndForwardOld(GarbageCollector *gc, Header *old)
{
  assert(!old->forwardingPointer);
  assert(Space_includes(&gc->oldSpace, old)
	 || Space_includes(&gc->oldSurvivorSpace, old)
	 || Space_includes(&gc->newSpace, old));
  Header *new;
  new= copyToSpace(old, &gc->oldSurvivorSpace);
  old->forwardingPointer= hdr2ptr(new);
  new->old= 1;
  assert(!new->forwardingPointer);
  assert(Space_includes(&gc->oldSurvivorSpace, new));
}

static void copyAndForwardOldObject(GarbageCollector *gc, GC_PTR ptr)
{
  if (GC_IS_POINTER(ptr)) {
    Header *hdr= ptr2hdr(ptr);
    if (!hdr->forwardingPointer) copyAndForwardOld(gc, hdr);
  }
}

static void scavengeOldReferentsOf(GarbageCollector *gc, Header *referrer)
{
  long map= referrer->map;
  if (map) {
    GC_PTR *optr;
    GC_PTR *next= (GC_PTR *)hdrNext(referrer, referrer->size);
    for (optr= referrer->contents;  (map & 1) && (optr < next);  ++optr, map >>= 1) {
      if (map & 1) {
	GC_PTR ptr= *optr;
	if (GC_IS_POINTER(ptr)) {
	  Header *referent= ptr2hdr(*optr);
	  if (!referent->forwardingPointer) copyAndForwardOld(gc, referent);
	  *optr= referent->forwardingPointer;
	}
      }
    }
  }
}

static void forwardOld(GC_PTR *optr)
{
  GC_PTR ptr= *optr;
  if (GC_IS_POINTER(ptr)) {
    Header *hdr= ptr2hdr(ptr);
    assert(hdr->forwardingPointer);
    *optr= hdr->forwardingPointer;
    assert(ptr2hdr(hdr->forwardingPointer)->old);
    assert(!ptr2hdr(hdr->forwardingPointer)->forwardingPointer);
  }
}

GC_API void GC_collect_all(GarbageCollector *gc)
{
  ++gc->full;
  ++gc->seq;

  sanityCheck(gc);

  if (Space_size(&gc->oldSurvivorSpace) <= Space_used(&gc->oldSpace) + Space_used(&gc->newSpace))
      Space_grow(&gc->oldSurvivorSpace, Space_used(&gc->oldSpace) + Space_used(&gc->newSpace) + GC_MIN_NEW_SPACE);

  assert(Space_size(&gc->oldSpace) <= Space_size(&gc->oldSurvivorSpace));

#if 0
  gc->callback(gc);
#endif

  gc->oldSurvivorSpace.next= gc->oldSurvivorSpace.base;

  int i;
  Header *scan;

  for (i= 0;  i < gc->rememberedSet.count;  ++i) ptr2hdr(gc->rememberedSet.at[i])->remembered= 0;
  gc->rememberedSet.count= 0;

  for (i= 0;  i < gc->roots.count;  ++i) copyAndForwardOldObject(gc, *(GC_PTR *)gc->roots.at[i]);
  for (i= 0;  i < gc->stack.count;  ++i) copyAndForwardOldObject(gc, *(GC_PTR *)gc->stack.at[i]);

  for (scan= gc->oldSurvivorSpace.base;  scan != gc->oldSurvivorSpace.next;  scan= hdrNext(scan, scan->size))
    scavengeOldReferentsOf(gc, scan);

  for (i= 0;  i < gc->roots.count;  ++i) forwardOld(gc->roots.at[i]);
  for (i= 0;  i < gc->stack.count;  ++i) forwardOld(gc->stack.at[i]);

  gc->newSpace.next= gc->newSpace.base;

  bzero(gc->newSpace.next, (void *)gc->newSpace.limit - (void *)gc->newSpace.next);

  {
    struct Space t= gc->oldSpace;
    gc->oldSpace= gc->oldSurvivorSpace;
    gc->oldSurvivorSpace= t;
  }

  if (Space_size(&gc->oldSurvivorSpace) < Space_size(&gc->oldSpace)) Space_grow(&gc->oldSurvivorSpace, Space_size(&gc->oldSpace));

#if (GC_INFO)
  printf("[GC %ld+%ld %2ld roots %2ld stack %4ld remembered,  %7ld / %7ld old, %7ld / %7ld new]\n",
	 (long)gc->full, (long)(gc->seq - gc->full),
	 (long)gc->roots.count, (long)gc->stack.count, (long)gc->rememberedSet.count,
	 (long)Space_used(&gc->oldSpace), (long)Space_size(&gc->oldSpace),
	 (long)Space_used(&gc->newSpace), (long)Space_size(&gc->newSpace)
    );

  memset(gc->oldSpace.next,	    -1, Space_free(&gc->oldSpace));
  memset(gc->oldSurvivorSpace.base, -2, Space_size(&gc->oldSurvivorSpace));
#endif

  sanityCheck(gc);
}

static void forward(GC_PTR *optr)
{
  GC_PTR ptr= *optr;
  if (GC_IS_POINTER(ptr)) {
    Header *hdr= ptr2hdr(ptr);
    if (hdr->forwardingPointer) {
      assert(!hdr->old);
      //printf("REMAP %p -> %p\n", ptr, hdr->forwardingPointer);
      *optr= hdr->forwardingPointer;
    }
  }
}

static void copyAndForward(GarbageCollector *gc, Header *old)
{
  assert(!old->forwardingPointer);
  Header *new;
  if (old->age < GC_MAX_AGE) {
    new= copyToSpace(old, &gc->survivorSpace);
    ++new->age;
#  if (GC_STATS)
    gc->totalAge += new->age;
    ++gc->totalNew;
#  endif
  }
  else {
    new= copyToSpace(old, &gc->oldSpace);
    new->old= 1;
    remember(gc, hdr2ptr(new));
  }
  old->forwardingPointer= hdr2ptr(new);
}

static void copyAndForwardObject(GarbageCollector *gc, GC_PTR ptr)
{
  if (GC_IS_POINTER(ptr)) {
    Header *hdr= ptr2hdr(ptr);
    if (!hdr->old && !hdr->forwardingPointer) {
      assert(!hdr->old);
      copyAndForward(gc, hdr);
      //printf("COPY FWD %p -> %p\n", ptr, hdr->forwardingPointer);
    }
  }
}

static int scavengeReferentsOf(GarbageCollector *gc, Header *referrer)
{
  int foundNewReferent= 0;
  long map= referrer->map;
  if (map) {
    GC_PTR *optr;
    GC_PTR *next= (GC_PTR *)hdrNext(referrer, referrer->size);
    for (optr= referrer->contents;  (map & 1) && (optr < next);  ++optr, map >>= 1) {
      if (map & 1) {
	GC_PTR ptr= *optr;
	if (GC_IS_POINTER(ptr)) {
	  Header *referent= ptr2hdr(*optr);
	  if (!referent->old) {
	    foundNewReferent= 1;
	    if (!referent->forwardingPointer) copyAndForward(gc, referent);
	    *optr= referent->forwardingPointer;
	  }
	}
      }
    }
  }
  return foundNewReferent;
}

GC_API void GC_collect(GarbageCollector *gc)
{
  sanityCheck(gc);
      
#if 0
  gc->callback(gc);
#endif

  if (Space_size(&gc->survivorSpace) < Space_size(&gc->newSpace)) Space_grow(&gc->survivorSpace, Space_size(&gc->newSpace));

  assert(Space_size(&gc->newSpace) == Space_size(&gc->survivorSpace));

//gc->count= 0;
#if (GC_STATS)
  struct timeval then;  gettimeofday(&then, 0);
#endif
  ++gc->seq;

#if !defined(NDEBUG)
  static char *cursors= "-/|\\";
  static int ix= 0;
  printf("7%c8", cursors[ix++/4%4]);  fflush(stdout);
#endif

  gc->survivorSpace.next= gc->survivorSpace.base;

  {
    int i;
    for (i= 0;  i < gc->roots.count;  ++i) copyAndForwardObject(gc, *(GC_PTR *)gc->roots.at[i]);
    for (i= 0;  i < gc->stack.count;  ++i) copyAndForwardObject(gc, *(GC_PTR *)gc->stack.at[i]);
  }

  {
    int     count= 0;
    Header *scan= gc->survivorSpace.base;

#  if (GC_STATS)
    gc->totalAge= gc->totalNew= 0;
#  endif

    for (;;) {
      int source;
      for (source= count;  source < gc->rememberedSet.count;  ++source)
	if (scavengeReferentsOf(gc, ptr2hdr(gc->rememberedSet.at[source]))) {
	  if (count == gc->rememberedSet.size) Table_grow(&gc->rememberedSet);
	  gc->rememberedSet.at[count++]= gc->rememberedSet.at[source];
	}
	else
	  ptr2hdr(gc->rememberedSet.at[source])->remembered= 0;
      gc->rememberedSet.count= count;
      if (scan == gc->survivorSpace.next) break;
      for (; scan != gc->survivorSpace.next;  scan= hdrNext(scan, scan->size))
	scavengeReferentsOf(gc, scan);
      if (count == gc->rememberedSet.count) break;
    }
  }

  {
    int i;
    for (i= 0;  i < gc->roots.count;  ++i) forward(gc->roots.at[i]);
    for (i= 0;  i < gc->stack.count;  ++i) forward(gc->stack.at[i]);
  }

  {
    struct Space t= gc->newSpace;
    gc->newSpace= gc->survivorSpace;
    gc->survivorSpace= t;
  }

  bzero(gc->newSpace.next, (void *)gc->newSpace.limit - (void *)gc->newSpace.next);

  gc->survivorSpace.next= gc->survivorSpace.base + (gc->newSpace.next - gc->newSpace.base);

#if (GC_INFO)
  //printf("7[H");
  printf("[GC %ld+%ld %2ld roots %2ld stack %4ld remembered,  %7ld / %7ld old, %7ld / %7ld new",
	 (long)gc->full, (long)(gc->seq - gc->full),
	 (long)gc->roots.count, (long)gc->stack.count, (long)gc->rememberedSet.count,
	 (long)Space_used(&gc->oldSpace), (long)Space_size(&gc->oldSpace),
	 (long)Space_used(&gc->newSpace), (long)Space_size(&gc->newSpace)
    );
#endif
#if (GC_STATS)
  struct timeval now, ingc;  gettimeofday(&now, 0);
  timersub(&now, &then, &ingc);
  timeradd(&ingc, &gc->gcTime, &gc->gcTime);
#  if (GC_INFO)
  long us= 0, msGC= 0, msTotal= 0;
  timersub(&now, &gc->startTime, &now);
  us= ingc.tv_sec * 1000000 + ingc.tv_usec;
  msGC= gc->gcTime.tv_sec * 1000 + gc->gcTime.tv_usec / 1000;
  msTotal= now.tv_sec * 1000 + now.tv_usec / 1000;
  if (us > 5000) printf(" %4ldms", us/1000); else printf(" %4ldus", us);
  printf(" %ld/%ldms total", msGC, msTotal);
  printf(" %4.1f%%", (float)msGC * 100.0 / (float)msTotal);
  if (GC_MAX_AGE) printf(" %.1f age", (gc->totalNew ? (float)gc->totalAge / (float)gc->totalNew : 0.0));
#  endif
#endif
#if (GC_INFO)
  printf("]\n");
  //printf("[K8");
  memset(gc->survivorSpace.base, -3, (void *)gc->survivorSpace.limit - (void *)gc->survivorSpace.base);
#endif
  sanityCheck(gc);
}

GC_API GC_PTR GC_malloc_mapped(GarbageCollector *gc, size_t n, long map)
{
  size_t sz= sizeof(Header) + ((n + (GC_ALIGN - 1)) & -GC_ALIGN);
  Header *next= hdrNext(gc->newSpace.next, sz);
#if (GC_TEST)
  if (1) {
#else
  if (next >= gc->newSpace.limit || gc->rememberedSet.count > GC_MAX_REMEMBERED /*|| ++gc->count > 1200000*/) {
#endif
    do {
      GC_collect(gc);
    } while (gc->rememberedSet.count >= GC_MAX_REMEMBERED);
    next= hdrNext(gc->newSpace.next, sz);
    if (next >= gc->newSpace.limit || (Space_used(&gc->oldSpace) >= Space_size(&gc->oldSurvivorSpace) - Space_size(&gc->newSpace))) {
      GC_collect_all(gc);
      assert(gc->newSpace.base == gc->newSpace.next);
      assert(gc->rememberedSet.count == 0);
      if (sz * 2 >= Space_size(&gc->newSpace)) {
#if (GC_INFO)
	printf("[GC grow new space to %ld]\n", sz * 2);
#endif
	Space_grow(&gc->newSpace, sz * 2);
	bzero(gc->newSpace.base, Space_size(&gc->newSpace));
      }
      next= hdrNext(gc->newSpace.next, sz);
    }
    if (Space_used(&gc->oldSpace) >= Space_size(&gc->oldSurvivorSpace) - Space_size(&gc->newSpace)) {
#if (GC_INFO)
      printf("[GC grow old space to %ld]\n", Space_size(&gc->oldSpace) + Space_size(&gc->newSpace));
#endif
      Space_grow(&gc->oldSurvivorSpace, Space_size(&gc->oldSpace) + Space_size(&gc->newSpace));
      GC_collect_all(gc);
      next= hdrNext(gc->newSpace.next, sz);
    }
  }
  assert(next == hdrNext(gc->newSpace.next, sz));
  Header *hdr= gc->newSpace.next;
  gc->newSpace.next= next;
  hdr->size= sz;
  hdr->map=  map;
  sanityCheck(gc);
#if (GC_SPY)
  static int spies= 0;
  if (spies++ % spyCount == 0) GC_update_spy(gc);
#endif

  return hdr2ptr(hdr);
}
  
GC_API GC_PTR GC_malloc(GarbageCollector *gc, size_t size)		{ return GC_malloc_mapped(gc, size, GC_POINTERS); }
GC_API GC_PTR GC_malloc_atomic(GarbageCollector *gc, size_t size)	{ return GC_malloc_mapped(gc, size, GC_ATOMIC); }

GC_API GC_PTR GC_check_store(GarbageCollector *gc, GC_PTR dst, GC_PTR src)
{
  if (GC_IS_POINTER(src)) {
    Header *dh= ptr2hdr(dst);
    Header *sh= ptr2hdr(src);
    if (!sh->old && dh->old && !dh->remembered) {
      dh->remembered= 1;
      Table_add(&gc->rememberedSet, dst);
    }
  }
  return src;
}

#if (GC_DEBUG)

static void sanityCheckPointer(GarbageCollector *gc, GC_PTR ptr)
{
  if (!GC_IS_POINTER(ptr)) return;
  Header *hdr= ptr2hdr(ptr);
  if (gc->oldSpace.base <= hdr && hdr < gc->oldSpace.next) {
  }
  else if (gc->newSpace.base <= hdr && hdr < gc->newSpace.next) {
  }
  else {
    fprintf(stderr, "pointer %p is not in object memory\n", ptr);
    abort();
  }
}

static void sanityCheckObject(GarbageCollector *gc, Header *hdr)
{
  assert(hdr->size >= sizeof(Header));
  long map= hdr->map;
  assert(!hdr->forwardingPointer);
  if (map) {
    GC_PTR *optr;
    GC_PTR *next= (GC_PTR *)hdrNext(hdr, hdr->size);
    if (GC_POINTERS == map)
      for (optr= hdr->contents;  optr < next;  ++optr) sanityCheckPointer(gc, *optr);
    else
      for (optr= hdr->contents;  (map & 1) && (optr < next);  ++optr, map >>= 1) if (map & 1) sanityCheckPointer(gc, *optr);
  }
}

static void sanityCheck(GarbageCollector *gc)
{
  Header *hdr;
  int i;
  for (i= 0;  i < gc->roots.count;  ++i) {
    GC_PTR ptr= *(GC_PTR *)gc->roots.at[i];
    sanityCheckPointer(gc, ptr);
  }
  for (i= 0;  i < gc->stack.count;  ++i) {
    GC_PTR ptr= *(GC_PTR *)gc->stack.at[i];
    sanityCheckPointer(gc, ptr);
  }
  for (i= 0;  i < gc->rememberedSet.count;  ++i) {
    GC_PTR ptr= (GC_PTR)gc->rememberedSet.at[i];
    sanityCheckPointer(gc, ptr);
    hdr=ptr2hdr(ptr);
    assert(hdr->old);
    assert(hdr->remembered);
  }
  for (hdr= gc->oldSpace.base;  hdr != gc->oldSpace.next;  hdr= hdrNext(hdr, hdr->size)) {
    assert(hdr->old);
    assert(!hdr->forwardingPointer);
    sanityCheckObject(gc, hdr);
    if (hdr->remembered) {
      GC_PTR ptr= hdr2ptr(hdr);
      for (i= 0;  i < gc->rememberedSet.count;  ++i) {
	GC_PTR rem= (GC_PTR)gc->rememberedSet.at[i];
	if (ptr == rem) break;
      }
      if (i == gc->rememberedSet.count) {
	fprintf(stderr, "pointer %p is remembered but not in set\n", ptr);
	abort();
      }
    }
  }
  for (hdr= gc->newSpace.base;  hdr != gc->newSpace.next;  hdr= hdrNext(hdr, hdr->size)) {
    assert(!hdr->old);
    assert(!hdr->forwardingPointer);
    assert(!hdr->remembered);
    sanityCheckObject(gc, hdr);
  }
}

#endif
